#include "Timer.mqh"
#property library

Timer::Timer(int _period,datetime _startTime=0) {
   period=_period;
   startTime=_startTime;
   ts=NULL;
   next=NULL;
   shedTime=0;
}
Timer::~Timer() {
   if (ts!=NULL) ts.Del(GetPointer(this));
}
bool Timer::Clock() {
   return false;
}
void Timer::OnAttach(void) {
}
void Timer::OnDetach(void) {
}
int Timer::Lag(void) {
   if (ts==NULL) return 0;
   if (ts.now<shedTime) return 0;
   return (ts.now-shedTime)%period;
}
int Timer::Omit(void) {
   if (ts==NULL) return 0;
   if (ts.now<shedTime) return 0;
   return (ts.now-shedTime)/period;
}
datetime TimeService::ServerTime() {
   return TimeLocal()-localOffset+serverOffset;
}
TimeService::TimeService(void) {
   first=NULL;
 
   serverOffset=TimeGMT()-TimeCurrent();  // GMT   
   localOffset=TimeGMT()-TimeLocal();     // GMT   

   if (IsTesting()) {
      now=TimeCurrent();
      serverOffset=0;
      localOffset=0;
   } else {
      now=ServerTime();
   }
}
TimeService::~TimeService() {
   while(first!=NULL) Del(first);
}
bool TimeService::Add(Timer *tptr) {
   if (tptr==NULL) return false;
   if (tptr.ts!=NULL) return false;
   tptr.ts=GetPointer(this);
   tptr.next=NULL;
   Shed(tptr);
   tptr.OnAttach();
   return true;
}
bool TimeService::Del(Timer *tptr) {
   if (tptr==NULL) return false;
   if (tptr.ts!=GetPointer(this)) return false;
   //          
   if (first==tptr) {
      first=tptr.next;
      tptr.next=NULL;
      tptr.ts=NULL;
      tptr.OnDetach();
      return true;
   } else {
      Timer *prev;
      for(prev=first;prev.next!=NULL;prev=prev.next) {
         if (prev.next==tptr) {
            prev.next=tptr.next;
            tptr.next=NULL;
            tptr.ts=NULL;
            tptr.OnDetach();
            return true;
         }
      }
   }
   return false;
}
void TimeService::Shed(Timer *tptr) {
   if (tptr==NULL) return;
   if (tptr.startTime==0) {
      tptr.startTime=now;
      tptr.shedTime=tptr.startTime+tptr.period;
   } else if (tptr.startTime>now) {
      tptr.shedTime=tptr.startTime;
   } else {
      tptr.shedTime=tptr.startTime+((now-tptr.startTime)/tptr.period+1)*tptr.period;
   }
   if (first==NULL) {
      first=tptr;
      tptr.next=NULL;
   } else if (first.shedTime>tptr.shedTime) {
      tptr.next=first;
      first=tptr;
   } else {
      Timer *prev;
      for(prev=first;prev.next!=NULL && prev.next.shedTime<=tptr.shedTime;) {
         prev=prev.next;
      }
      tptr.next=prev.next;
      prev.next=tptr;
   }
}
void TimeService::Clock() {
   Timer *timer;
   now=ServerTime();
   if (first==NULL || first.shedTime>now) return;  //       
   // ToDo -       ,     ,     - 
   while(first!=NULL && first.shedTime<=now) {
      timer=first;
      first=timer.next;
      timer.next=NULL;
      if (timer.Clock()) Shed(timer);
      else {
         timer.ts=NULL;
         timer.OnDetach();
      }
   }
}

